/*******************************************************************************
 * Author: Jérôme Dubois       Jerome.Dubois@ensimag.imag.fr jerome@stud.ntnu.no
 * Created: 2008-01-15
 * File  : 2d_polyphase_filter.c
 * Copyright 2008 Jerome Dubois
 *
 * Goal: Compute a 2D polyphase filter. Signals are 2D signals.
 ******************************************************************************/

//#include <string.h> // strlen, strncpy
#include <stdlib.h> // exit, EXIT_SUCCESS
#include <stdio.h> // printf, FILE. Already included in grim.h
// #include <stdint.h>
// #include <math.h> // sin, sqrt, pow. Needs linking with option -lm
 #include <assert.h> // assert
#include <time.h>
#include "filters.h"
#include "stopwatch.h" // Timer stuff: code around gettimeofday() of sys/time.h

// Function declaration
extern int polyphase_filter_2d(fgrim *in, fgrim *out, fgrim *h0_coef); // here
// defined in a_my_convolutionSeparable.cu
extern int filter_periodic_horizontalGPU(fgrim *fg, fgrim *coef, int offset);
extern int filter_periodic_verticalGPU(  fgrim *fg, fgrim *coef, int offset);

#define IMAGES "../images/"   // path to the images (to read and create)
#define noGPU
#define TIMER
#define SAVE_IMAGES

#define DEFINED
#undef  UNDEFINED

#ifdef GPU
#include <cuda_runtime.h> // for cudaMallocHost
#endif

/*############################################################################*/
/*#                      various MACROS                                      #*/
/*############################################################################*/

#ifdef SAVE_IMAGES
#define SAVE(sig,file) fgrim_write_bmp((sig),(IMAGES file))
#else
#define SAVE(sig,file)
#endif

// Timer 1: clock() of time.h
#define TIMER_1_DECLARE clock_t start, stop; float time;
#define TIMER_1_BEGIN   start = clock();
#define TIMER_1_END(direction)                                  \
   stop = clock();                                              \
   time = ((float)stop - start) / CLOCKS_PER_SEC * 1000;        \
   printf( direction " filtering of both bands: Timer 1: %f ms\n", time);
#define TIMER_1_DELETE

// Timer 2: gettimeofday() of sys/time.h through stopwatch.h
#define TIMER_2_DECLARE float stopwatch_time;
#define TIMER_2_BEGIN   stopwatch_reset(); stopwatch_start();
#define TIMER_2_END(direction)                                  \
    stopwatch_stop();                                           \
    stopwatch_time = stopwatch_getTime();                       \
    printf( direction " filtering of both bands: Timer 2: %6.4f ms\n", stopwatch_time );
#define TIMER_2_DELETE

#ifdef TIMER
#define TIMER_DECLARE        TIMER_1_DECLARE        TIMER_2_DECLARE 
#define TIMER_BEGIN          TIMER_1_BEGIN          TIMER_2_BEGIN  // 2 has prio
#define TIMER_END(direction) TIMER_2_END(direction) TIMER_1_END(direction) // id
#define TIMER_DELETE         TIMER_1_DELETE         TIMER_2_DELETE
#else
#define TIMER_DECLARE
#define TIMER_BEGIN
#define TIMER_END(direction)
#define TIMER_DELETE
#endif

#ifdef GPU
#define MACRO_filter_periodic_horizontal(signal,coef,offset) \
              filter_periodic_horizontalGPU((signal), (coef), (offset));
#define MACRO_filter_periodic_vertical(  signal,coef,offset) \
              filter_periodic_verticalGPU(  (signal), (coef), (offset));
#else
#define MACRO_filter_periodic_horizontal(signal,coef,offset) \
              filter_periodic_horizontal((signal), (coef), (offset));
#define MACRO_filter_periodic_vertical(  signal,coef,offset) \
              filter_periodic_vertical(  (signal), (coef), (offset));
#endif

/*############################################################################*/
/*#                      various functions                                   #*/
/*############################################################################*/
#ifdef GPU
int cudaMalloc_fgrim_create(fgrim *fimg, int width, int height);
int cudaMalloc_fgrim_free(fgrim *fimg);

int cudaMalloc_fgrim_create(fgrim *fimg, int width, int height){
  fimg->w = width;
  fimg->h = height;
  cudaMallocHost( (void**) &fimg->pix, width * height * sizeof(float) );
  return EXIT_SUCCESS;
}
int cudaMalloc_fgrim_free(fgrim *fimg){
  fimg->w = 0;
  fimg->h = 0;
  cudaFreeHost(fimg->pix); 
  return EXIT_SUCCESS;
}  
#endif // #ifdef GPU

/*############################################################################*/
/*#                      2d_polyphase_filter                                 #*/
/*############################################################################*/
int polyphase_filter_2d(fgrim *in, fgrim *out, fgrim *h0_coef){
  int i, offset1, offset2;
  int w, w2; // width of the signals
  int fs, fs2; // size of the filters
  fgrim p0_coef, p1_coef;   // polyphase filter coefficients
  TIMER_DECLARE;

    /* preparation */
    w  = in->w;
    fs = h0_coef->w;
    assert(0 ==  w%2);
    assert(0 == fs%2);
    assert(w == in->h);
    w2  = w /2;
    fs2 = fs/2;
    
    // coefficient preparation
    fgrim_create(&p0_coef, fs2, 1);
    fgrim_create(&p1_coef, fs2, 1);
    for (i=0; i<fs2; i++){
      p0_coef.pix[i] = h0_coef->pix[2*i  ];
      p1_coef.pix[i] = h0_coef->pix[2*i+1];
    }
    offset1 = 0; // assert( abs( offset1)<w ) ?
    offset2 = fs2-offset1-1; // assert( abs( offset1)<w ) ?

    // bank preparation
    fgrim *sA1, sA2, *sA13a14;
    fgrim sA4a5b1, sA6a7b1, sA8b1, sA9a10b1, sA11a12b1; // sA4a5b1: signal A4 and A5 band 1
    fgrim sA4a5b2, sA6a7b2, sA8b2, sA9a10b2, sA11a12b2; // sA8b2  : signal A8 band 2
        
    sA1 = in;  // pointer copy
    fgrim_create(&sA2, w, w);
    //fgrim_create(&sA4a5b1  , w , w2); fgrim_create(&sA4a5b2  , w , w2);
    fgrim_create(&sA8b1    , w2, w ); fgrim_create(&sA8b2    , w2, w );
#ifdef GPU
    //cudaError_t err;  // err = cudaMalloc...
    cudaMalloc_fgrim_create(&sA4a5b1  , w , w2); cudaMalloc_fgrim_create(&sA4a5b2  , w , w2);
    //printf( cudaGetErrorString( err ) ); printf("\n");
    cudaMalloc_fgrim_create(&sA6a7b1  , w2, w ); cudaMalloc_fgrim_create(&sA6a7b2  , w2, w );
    //cudaMalloc_fgrim_create(&sA9a10b1 , w2, w ); cudaMalloc_fgrim_create(&sA9a10b2 , w2, w );
    //cudaMalloc_fgrim_create(&sA11a12b1, w , w2); cudaMalloc_fgrim_create(&sA11a12b2, w , w2);
    fgrim_create(&sA9a10b1 , w2, w ); fgrim_create(&sA9a10b2 , w2, w );
    fgrim_create(&sA11a12b1, w , w2); fgrim_create(&sA11a12b2, w , w2);
#else
    fgrim_create(&sA4a5b1  , w , w2); fgrim_create(&sA4a5b2  , w , w2);
    fgrim_create(&sA6a7b1  , w2, w ); fgrim_create(&sA6a7b2  , w2, w );
    fgrim_create(&sA9a10b1 , w2, w ); fgrim_create(&sA9a10b2 , w2, w );
    fgrim_create(&sA11a12b1, w , w2); fgrim_create(&sA11a12b2, w , w2);
#endif
    sA13a14 = out; // pointer copy
    fgrim_create(sA13a14, w, w);


    /* analysis */
  SAVE(sA1, "sA01.bmp");
    fgrim_copy(sA1, &sA2); // Not optimal, but we don't want to modify sA1
    frequency_shift_PI(&sA2);
  SAVE(&sA2, "sA02.bmp");
    downsample_quincunx_q2(&sA2, &sA4a5b1, &sA4a5b2);
  SAVE(&sA4a5b1, "sA04b1.bmp"); SAVE(&sA4a5b2, "sA04b2.bmp");
    TIMER_BEGIN;
      MACRO_filter_periodic_horizontal(&sA4a5b1, &p0_coef, offset1);
      MACRO_filter_periodic_horizontal(&sA4a5b2, &p1_coef, offset1);
    TIMER_END("horizontal")
  SAVE(&sA4a5b1, "sA05b1.bmp"); SAVE(&sA4a5b2, "sA05b2.bmp");
    reorganize_h2v(&sA4a5b1, &sA6a7b1);
    reorganize_h2v(&sA4a5b2, &sA6a7b2);
  SAVE(&sA6a7b1, "sA06b1.bmp"); SAVE(&sA6a7b2, "sA06b2.bmp");
    TIMER_BEGIN;
      MACRO_filter_periodic_vertical(&sA6a7b1, &p0_coef, offset1);
      MACRO_filter_periodic_vertical(&sA6a7b2, &p1_coef, offset1);
    TIMER_END("vertical  ")
  SAVE(&sA6a7b1, "sA07b1.bmp"); SAVE(&sA6a7b2, "sA07b2.bmp");
    polyphase_combine(&sA6a7b1, &sA6a7b2, &sA8b1, &sA8b2);
  SAVE(&sA8b1, "sA08b1.bmp"); SAVE(&sA8b2, "sA08b2.bmp");


    /* processing */
    //fgrim_set_color(&sA8b1, BLACK);
    //fgrim_set_color(&sA8b2, BLACK);
    //SAVE(&sA8b1, "sA08b1bis.bmp"); SAVE(&sA8b2, "sA08b2bis.bmp");
  

    /* synthesis */
    polyphase_combine(&sA8b1, &sA8b2, &sA9a10b1, &sA9a10b2);
  SAVE(&sA9a10b1, "sA09b1.bmp"); SAVE(&sA9a10b2, "sA09b2.bmp");
    filter_periodic_vertical(&sA9a10b1, &p1_coef, offset2);
    filter_periodic_vertical(&sA9a10b2, &p0_coef, offset2);
    //filter_periodic_verticalGPU(&sA9a10b1, &p1_coef, offset2); // no offset
    //filter_periodic_verticalGPU(&sA9a10b2, &p0_coef, offset2); // no offset
  SAVE(&sA9a10b1, "sA10b1.bmp"); SAVE(&sA9a10b2, "sA10b2.bmp");
    reorganize_v2h(&sA9a10b1, &sA11a12b1);
    reorganize_v2h(&sA9a10b2, &sA11a12b2);
    amplify(&sA11a12b1, 2);
    amplify(&sA11a12b2, 2);
  SAVE(&sA11a12b1, "sA11b1.bmp"); SAVE(&sA11a12b2, "sA11b2.bmp");
    filter_periodic_horizontal(&sA11a12b1, &p1_coef, offset2);
    filter_periodic_horizontal(&sA11a12b2, &p0_coef, offset2);
    //filter_periodic_horizontalGPU(&sA11a12b1, &p1_coef, offset2);
    //filter_periodic_horizontalGPU(&sA11a12b2, &p0_coef, offset2);
    amplify(&sA11a12b1, 4);
    amplify(&sA11a12b2, 4);
  SAVE(&sA11a12b1, "sA12b1.bmp"); SAVE(&sA11a12b2, "sA12b2.bmp");
    upsample_quincunx_q2(sA13a14, &sA11a12b1, &sA11a12b2);
  SAVE(sA13a14, "sA13.bmp");
    frequency_shift_PI(sA13a14);
  SAVE(sA13a14, "sA14.bmp");
  SAVE(sA13a14, "sA00.bmp");
  float diff = fgrim_diff(sA1, sA13a14);
  printf("diff=%f percent\n", diff );
  

    /* ending */
    sA1  = NULL;
    fgrim_free(&sA2);
    //fgrim_free(&sA4a5b1  ); fgrim_free(&sA4a5b2  );
    fgrim_free(&sA8b1    ); fgrim_free(&sA8b2    );
#ifdef GPU
    cudaMalloc_fgrim_free(&sA4a5b1  ); cudaMalloc_fgrim_free(&sA4a5b2  );
    cudaMalloc_fgrim_free(&sA6a7b1  ); cudaMalloc_fgrim_free(&sA6a7b2  );
    //cudaMalloc_fgrim_free(&sA9a10b1 ); cudaMalloc_fgrim_free(&sA9a10b2 );
    //cudaMalloc_fgrim_free(&sA11a12b1); cudaMalloc_fgrim_free(&sA11a12b2);
    fgrim_free(&sA9a10b1 ); fgrim_free(&sA9a10b2 );
    fgrim_free(&sA11a12b1); fgrim_free(&sA11a12b2);
#else
    fgrim_free(&sA4a5b1  ); fgrim_free(&sA4a5b2  );
    fgrim_free(&sA6a7b1  ); fgrim_free(&sA6a7b2  );
    fgrim_free(&sA9a10b1 ); fgrim_free(&sA9a10b2 );
    fgrim_free(&sA11a12b1); fgrim_free(&sA11a12b2);
#endif
    sA13a14 = NULL;
    fgrim_free(&p0_coef); fgrim_free(&p1_coef);
    TIMER_DELETE;

  return EXIT_SUCCESS;
}

