/*******************************************************************************
 * Author : Jérôme Dubois       Jerome.Dubois@ensimag.imag.fr jerome@stud.ntnu.no
 * Created: 2007-11-5
 * File  : dfb.c
 * Copyright 2007 Jerome Dubois
 *
 * Goal:   Serial implementation of an image directional decomposition algorithm
 *         inspired by Mark J.T. Smith's work.
 ******************************************************************************/

#include <string.h> // strlen, strncpy
#include <stdlib.h> // exit, EXIT_SUCCESS
// <stdio.h> // printf. Already included in grim.h
// #include <stdint.h>
// #include <math.h> // sin, sqrt, pow. Needs linking with option -lm
// #include <assert.h> // assert
#include <getopt.h> // getopt
#include "filters.h"
#include "dprintf.h"

// Function declaration
extern void help(char* cmd);
extern void error(char* cmd, char* msg);
extern int polyphase_filter_2d(fgrim *in, fgrim *out, fgrim *h0_coef); // in 2d_polyphase_filter.c
extern void dfb_test(void); // defined in dfb_test.c

#define ARG_SIZE 128 // max size of string arguments (including last 0x0)
static int verbose;


/*############################################################################*/
/*#                      functions                                           #*/
/*############################################################################*/
void help(char* cmd){
  printf("Usage: %s [options] input_image\n",cmd);
  printf("Options:\n"
  "  -h, --help               print help\n"
  "  -o <file>                Place the output into <file>. Default is out.grim    \n"
  "  -v, --verbose            Print various information\n"
  "Author: Jerome.Dubois@ensimag.imag.fr\n"
  "Copyright 2007 2008 Jerome Dubois\n");
}
void error(char* cmd, char* msg) {
    fprintf(stderr, "Error: %s\n",msg);
    fprintf(stderr, "Use '%s -help' for help\n", cmd);
    exit(EXIT_FAILURE);
}

/*############################################################################*/
/*#                      main                                                #*/
/*############################################################################*/
int main(int argc, char * argv[]){
  char c;
  fgrim filter_coef;
  grim   grim_in,  grim_out;
  fgrim fgrim_in, fgrim_out;

  // Arguments
//  extern char *optarg; extern int optind;  // in getopt.h
  char output_file_name[ARG_SIZE] = "out.grim";
  char input_image_name[ARG_SIZE] ;
  FILE* fp_output_file;

  /* Command line analysis */

  // Get options
  verbose = FALSE;
  while ((c = getopt(argc, argv, "ho:v")) != -1) {
      switch(c) {
      case 'h': help(argv[0]) ; return EXIT_SUCCESS; break;
      case 'o':
        if (strlen(optarg) >= ARG_SIZE) error(argv[0], "Output file name is too long");
        strncpy(output_file_name,optarg,ARG_SIZE);
        if (output_file_name[0] == '-') error(argv[0],"Bad output file name");
        break;
      case 'v': verbose = TRUE; break;
      default: error("Wrong command line parameters",argv[0]);
      }
  }

  // Check number of parameters
  if      ((argc - optind) <  0) error(argv[0],"internal error");
  else if ((argc - optind) == 0) error(argv[0],"missing arguments (probably input_image)");
  else if ((argc - optind) >  1) error(argv[0],"too many parameters");

  // input_image
  if (strlen(argv[optind]) >= ARG_SIZE) error(argv[0], "Input image name is too long");
  strncpy(input_image_name,argv[optind],ARG_SIZE);

  // manage and check input/output files
  grim_read_grim(input_image_name,&grim_in);
  if ( (fp_output_file = fopen(output_file_name,"w")) == NULL)
    error(argv[0], "Cannot write output file");
  fclose(fp_output_file);

  // check arguments
  d_printf("Parameters: Verbose: %s\n", verbose ? "yes" : "no" );
  d_printf("Parameters: Input image: %s\n", input_image_name   );
  d_printf("Parameters: Output file: %s\n", output_file_name   );
  
  /* Begin the real program */

  // tests
  //dfb_test(); exit(0);

  // init
  init_coef(&filter_coef);
  grim_to_fgrim(&grim_in, &fgrim_in);

  // signal processing
  polyphase_filter_2d(&fgrim_in, &fgrim_out, &filter_coef);
  
  // convert and export result
  fgrim_to_grim(&fgrim_out, &grim_out);
  grim_write_grim(&grim_out, output_file_name);

  // free
  free_coef(&filter_coef);
  fgrim_free(&fgrim_in); fgrim_free(&fgrim_out);
  grim_free(  &grim_in); grim_free(  &grim_out);

  return EXIT_SUCCESS;
}

