/*******************************************************************************
 * Author: Jérôme Dubois       Jerome.Dubois@ensimag.imag.fr jerome@stud.ntnu.no
 * File  : dfb_cuda.cu
 * Copyright 2007 Jerome Dubois
 *
 * The program uses some code from the NVIDIA CUDA SDK 1.1 sample
 * convolutionSeparable. The license authorizes us to use it and deny all
 * liability. The NVIDIA copyright and license is at the end of the file.
 ******************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <cutil.h>
#include <time.h>
#include "grim.h"

#include <dfb_cuda_kernel.cu>

#define DEFINED
#undef  UNDEFINED

#ifdef UNDEFINED
#define TIMER_DECLARE        \
    double gpuTime;          \
    unsigned int hTimer;     \
    cutCreateTimer(&hTimer);
#define TIMER_BEGIN cutResetTimer(hTimer); cutStartTimer(hTimer) ;
#define TIMER_END(direction)                                                   \
    cutStopTimer(hTimer);                                                      \
    gpuTime = cutGetTimerValue(hTimer);                                        \
    printf("GPU " direction " convolution time : %f msec //%f Mpixels/sec\n",  \
           gpuTime, 1e-6 * data_w * data_h / (gpuTime * 0.001));
#define TIMER_DELETE  cutDeleteTimer(hTimer);
#else
#define TIMER_DECLARE
#define TIMER_BEGIN
#define TIMER_END(direction)
#define TIMER_DELETE
#endif

const int KERNEL_SIZE = KERNEL_W * sizeof(float);

////////////////////////////////////////////////////////////////////////////////
// Common host and device functions
////////////////////////////////////////////////////////////////////////////////
//Round a / b to nearest higher integer value
int iDivUp(int a, int b){
    return (a % b != 0) ? (a / b + 1) : (a / b);
}

//Round a / b to nearest lower integer value
int iDivDown(int a, int b){
    return a / b;
}

//Align a to nearest higher multiple of b
int iAlignUp(int a, int b){
    return (a % b != 0) ?  (a - a % b + b) : a;
}

//Align a to nearest lower multiple of b
int iAlignDown(int a, int b){
    return a - a % b;
}


////////////////////////////////////////////////////////////////////////////////
// Reference CPU convolution
////////////////////////////////////////////////////////////////////////////////
extern "C" void convolutionRowCPU(
    float *h_Result,
    float *h_Data,
    float *h_Kernel,
    int dataW,
    int dataH,
    int kernelR
);

extern "C" void convolutionColumnCPU(
    float *h_Result,
    float *h_Data,
    float *h_Kernel,
    int dataW,
    int dataH,
    int kernelR
);

extern "C" int filter_periodic_horizontalGPU(
    fgrim *fg,
    fgrim *coef,
    int offset
);

extern "C" int filter_periodic_verticalGPU(
    fgrim *fg,
    fgrim *coef,
    int offset
);


////////////////////////////////////////////////////////////////////////////////
// filter_periodic_horizontalGPU
////////////////////////////////////////////////////////////////////////////////
int filter_periodic_horizontalGPU(fgrim *fg, fgrim *coef, int offset){
    float *h_Kernel, *h_Data; // host kernel and input/output data
    float *d_DataI, *d_DataO; // device global memory: input and output data
    TIMER_DECLARE;
    
    int data_w = fg->w;
    int data_h = fg->h;
    int data_size = data_w * data_h * sizeof(float);

    /* Preparing data */
    h_Data = fg->pix;
    cudaMalloc( (void **)&d_DataI, data_size); // ~5 ms for both malloc
    cudaMalloc( (void **)&d_DataO, data_size); // and free for 2 calls

    /* Preparing filter coefficients */
    h_Kernel = coef->pix;

    /* Copy data and kernel to device */
    cudaMemcpyToSymbol(d_Kernel, h_Kernel, KERNEL_SIZE);
    cudaMemcpy(d_DataI, h_Data, data_size, cudaMemcpyHostToDevice);//2calls:20ms

    /* Prepare execution */
    dim3 blockGridRows(iDivUp(data_w, ROW_TILE_W), data_h);
    dim3 threadBlockRows(KERNEL_RADIUS_ALIGNED + ROW_TILE_W + KERNEL_RADIUS);

    /* GPU convolution */

    cudaThreadSynchronize() ;
    TIMER_BEGIN;
    convolutionRowGPU<<<blockGridRows, threadBlockRows>>>(
         d_DataO, // result
         d_DataI, // data
         data_w,
         data_h
    );
    cudaThreadSynchronize() ;
    TIMER_END("horizontal");

    /* Reading back GPU results */
    cudaMemcpy(h_Data, d_DataO, data_size, cudaMemcpyDeviceToHost);

    /* Shutting down */
    cudaFree(d_DataI);
    cudaFree(d_DataO);
    //free(h_Kernel);
    TIMER_DELETE;

    return(EXIT_SUCCESS);
}


////////////////////////////////////////////////////////////////////////////////
// filter_periodic_verticalGPU
////////////////////////////////////////////////////////////////////////////////
int filter_periodic_verticalGPU(fgrim *fg, fgrim *coef, int offset){
    float *h_Kernel, *h_Data;
    float *d_DataI, *d_DataO;
    TIMER_DECLARE;
    
    int data_w = fg->w;
    int data_h = fg->h;
    int data_size = data_w * data_h * sizeof(float);

    /* Preparing data */
     h_Data = fg->pix; // data
     cudaMalloc( (void **)&d_DataI, data_size);
     cudaMalloc( (void **)&d_DataO, data_size);

    /* Preparing filter coefficients */
    h_Kernel = coef->pix;

    /* Copy data and kernel to device */
    cudaMemcpyToSymbol(d_Kernel, h_Kernel, KERNEL_SIZE);
    cudaMemcpy(d_DataI, h_Data, data_size, cudaMemcpyHostToDevice);

    /* Prepare execution */
    dim3 blockGridColumns(iDivUp(data_w, COLUMN_TILE_W),
                          iDivUp(data_h, COLUMN_TILE_H));
    dim3 threadBlockColumns(COLUMN_TILE_W, 8);

    /* GPU convolution */
    cudaThreadSynchronize();
    TIMER_BEGIN;
    convolutionColumnGPU<<<blockGridColumns, threadBlockColumns>>>(
        d_DataO, // result
        d_DataI, // data
        data_w,
        data_h,
        COLUMN_TILE_W * threadBlockColumns.y, // shared mem stride
        data_w * threadBlockColumns.y         // global mem stride
    );
    cudaThreadSynchronize();
    TIMER_END("vertical  ");

    /* Reading back GPU results */
    cudaMemcpy(h_Data, d_DataO, data_size, cudaMemcpyDeviceToHost);

    /* Shutting down */
    cudaFree(d_DataI);
    cudaFree(d_DataO);
    //free(h_Kernel);
    TIMER_DELETE;

    return(EXIT_SUCCESS);
}

/******************************************************************************/
/* 
 * Some code used here is subject to NVIDIA ownership, so we must copy the
 * following disclaimer
 */
 
/*
 * Copyright 1993-2007 NVIDIA Corporation.  All rights reserved.
 *
 * NOTICE TO USER:   
 *
 * This source code is subject to NVIDIA ownership rights under U.S. and 
 * international Copyright laws.  Users and possessors of this source code 
 * are hereby granted a nonexclusive, royalty-free license to use this code 
 * in individual and commercial software.
 *
 * NVIDIA MAKES NO REPRESENTATION ABOUT THE SUITABILITY OF THIS SOURCE 
 * CODE FOR ANY PURPOSE.  IT IS PROVIDED "AS IS" WITHOUT EXPRESS OR 
 * IMPLIED WARRANTY OF ANY KIND.  NVIDIA DISCLAIMS ALL WARRANTIES WITH 
 * REGARD TO THIS SOURCE CODE, INCLUDING ALL IMPLIED WARRANTIES OF 
 * MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE.
 * IN NO EVENT SHALL NVIDIA BE LIABLE FOR ANY SPECIAL, INDIRECT, INCIDENTAL, 
 * OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS 
 * OF USE, DATA OR PROFITS,  WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE 
 * OR OTHER TORTIOUS ACTION,  ARISING OUT OF OR IN CONNECTION WITH THE USE 
 * OR PERFORMANCE OF THIS SOURCE CODE.  
 *
 * U.S. Government End Users.   This source code is a "commercial item" as 
 * that term is defined at  48 C.F.R. 2.101 (OCT 1995), consisting  of 
 * "commercial computer  software"  and "commercial computer software 
 * documentation" as such terms are  used in 48 C.F.R. 12.212 (SEPT 1995) 
 * and is provided to the U.S. Government only as a commercial end item.  
 * Consistent with 48 C.F.R.12.212 and 48 C.F.R. 227.7202-1 through 
 * 227.7202-4 (JUNE 1995), all U.S. Government End Users acquire the 
 * source code with only those rights set forth herein. 
 *
 * Any use of this source code in individual and commercial software must 
 * include, in the user documentation and internal comments to the code,
 * the above Disclaimer and U.S. Government End Users Notice.
 */
