/* Copyright 2007 2008 Jerome Dubois Jerome.Dubois@ensimag.imag.fr
 *
 * The aim of this file is to provide a solution to print some message only if
 * a specific macro (usually DEBUG) is defined, and do nothing otherwise.
 *
 * History:
 * 2007-xx-xx: creation
 * 2007-xx-xx: add of d_fprintf for stream output
 * 2008-02-10: searched the web for existing solution, added a summary of my
 *   research in comments. Add inline to my functions.
 *
 */

#include <stdarg.h> // va_list, va_start, vprintf,...
//#include <string.h>
#include <stdio.h> // printf, FILE... also needed for v*printf

#include "dprintf.h"

#ifdef DEBUG
#define TRACE
#endif

#ifdef TRACE
void d_printf(char *format,...)
{
  va_list ap;
  va_start(ap,format);
  vprintf(format,ap);
  va_end(ap);
}
void d_fprintf(FILE* stream, char *format,...)
{
  va_list ap;
  va_start(ap,format);
  vfprintf(stream,format,ap);
  va_end(ap);
}
#else
// Because they are inline, return void and are empty, the compiler should skip
// these functions. However, it does not (checked the assembly), even with -O3.
// The compiler may warn that the parameters are unused.
inline void d_printf(char *format,...)
{
}
inline void d_fprintf(FILE* stream, char *format,...)
{
}
#endif

// My initial solution (works well but returns int, which makes impossible to
// have a blank function for release)
/*
int d_printf(const char *format, ...)
{
  int ret=0;
#ifdef DEBUG
  va_list ap; // argument pointer
  va_start(ap,format);
  ret = vprintf(format,ap); // vprintf because we use a va_list instead of ...
  va_end(ap);
#endif
  return ret;
}
int d_fprintf(FILE* stream, const char *format, ...)
{
  int ret=0;
#ifdef DEBUG
  va_list ap;
  va_start(ap, format);
  ret = vfprintf(stream, format, ap);
#endif
  return ret;
}
*/

/* Solutions found on the web:

The page http://forum.qtfr.org/viewtopic.php?id=3650&p=1 (in French) presents
various methods for a printf in mode debug and nothing (no overhead) in mode
release. Most of them are for C++, but some of them can be adapted in C:
Solution 1:
  #include <stdio.h>
  #define MACROPRINTF(msg) printf msg
  int main()
  {
      MACROPRINTF(("j'ai %d ans", 30)); // WARNING: needs double parenthesis
  }
Solution 2: The fact that they are inlined, there is no overhead in release mode
  #ifdef TRACE
  inline void MYPRINTF(char *msg,...)
  {
    va_list ap;
    va_start(ap,msg);
    vprintf(msg,ap);
    va_end(ap);
  }
  #else
  inline void MYPRINTF(char *msg,...)
  {
  }
  #endif
  
A search of kewords "debug printf" on Exalead gives some results:
http://frog.raindrop.jp/knowledge/archives/000598.html (in Japanese)
  #ifdef	_DEBUG
  #define	debug_printf	printf
  #else	// _DEBUG
  #define	debug_printf	1? (void)0: printf
  #endif	// _DEBUG
http://tech.huisu.net/debug_printf.html
gcc accepts a variable number of arguments for macros:
  #ifdef DEBUG
  #define PRINTF(format, args...) printf(format , ## args)
  #else
  #define PRINTF(format, args...)
  #endif
Play with comma so that the first macro sees only 1 argument:
  #define DEBUG(args) (printf("DEBUG: "), printf(args))
  #define _ ,
  DEBUG("i = %d" _ i)
The page http://users.bestweb.net/~ctips/tip059.html proposes another approach
that can be use with other functions than printf. The basic idea of the call is:
  DBG( fprintf(stderr, "int_stack: pushed %d onto %p\n", val, stack));
The declaration depends of macros to en/dis-able printing in a per file manner.
*/




