/*******************************************************************************
 * Author: Jérôme Dubois       Jerome.Dubois@ensimag.imag.fr jerome@stud.ntnu.no
 * Copyright 2007 2008 Jerome Dubois
 *
 *
 ******************************************************************************/
#include <stdlib.h>
#include <assert.h>
#include <string.h> // memcpy
#include <time.h>
#include "filters.h"

/******************************************************************************
 *                       frequency_shift_PI
 * Goal:
 *   Apply a filter resulting in an horizontal shift by PI in the frequency
 *   domain.
 *   This shift result in the space domain by a multiplication by e^{i*PI*n}
 *   = (-1)^n with n=column number and here i is imaginative unit (i^2=-1)
 * Interface:
 *   fimg: [in-out] the fgrim to operate on
 * Return:
 *   status of execution
 ******************************************************************************/
int frequency_shift_PI(fgrim *fimg){
  int i,j;
  for (i=0; i<fimg->h; i++){ // raw i
    for (j=0; j<fimg->w; j++){ // column j
      if ((j)%2) fimg->pix[j + i*fimg->w] *= -1;
    }
  }
  return EXIT_SUCCESS;
}

/******************************************************************************
 *                       init_coef
 * Goal:
 *   Initialize the coefficients for the filters.
 * Return:
 *   status of execution
 ******************************************************************************/
/*
  ASGDIP: "A Study Guide for Digital Image Processing" (Early released edition)
          by Mark J. T. Smith and Alen Docef
*/

void init_coef(fgrim *filter_coef){
    assert( 0 == FILTER_LENGTH %2 ); // multiple of 2
    fgrim_create(filter_coef, FILTER_LENGTH, 1);
#if ( FILTER_LENGTH == 4 )
        fprintf(stdout, "Warning: This filter is ONLY for debugging purpose, and gives BAD result !\n");
        filter_coef->pix[ 0] =  1 ;
        filter_coef->pix[ 1] =  2 ;
        filter_coef->pix[ 2] =  2 ;
        filter_coef->pix[ 3] =  1 ;
#elif ( FILTER_LENGTH == 8 )
        // from ASGDIP page 218. "Table 5.5: QMF Filter Bank Coefficients." 8A
        filter_coef->pix[ 0] =  .93871500e-02 ; //   1
        filter_coef->pix[ 1] = -.70651830e-01 ; // - 7
        filter_coef->pix[ 2] =  .69428270e-01 ; //   7
        filter_coef->pix[ 3] =  .48998080e+00 ; //  48
        filter_coef->pix[ 4] =  .48998080e+00 ;
        filter_coef->pix[ 5] =  .69428270e-01 ;
        filter_coef->pix[ 6] = -.70651830e-01 ;
        filter_coef->pix[ 7] =  .93871500e-02 ;
#elif ( FILTER_LENGTH == 16 )
        // from ASGDIP page 218. "Table 5.5: QMF Filter Bank Coefficients." 16A
        filter_coef->pix[ 0] =  .10501670e-02 ; //  0.1
        filter_coef->pix[ 1] = -.50545260e-02 ; // -0.5
        filter_coef->pix[ 2] = -.25897560e-02 ; // -0.2
        filter_coef->pix[ 3] =  .27641400e-01 ; //  2.7
        filter_coef->pix[ 4] = -.96663760e-02 ; // -0.9
        filter_coef->pix[ 5] = -.90392230e-01 ; // -9
        filter_coef->pix[ 6] =  .97798170e-01 ; //  10
        filter_coef->pix[ 7] =  .48102840e+00 ; //  48
        filter_coef->pix[ 8] =  .48102840e+00 ;
        filter_coef->pix[ 9] =  .97798170e-01 ;
        filter_coef->pix[10] = -.90392230e-01 ;
        filter_coef->pix[11] = -.96663760e-02 ;
        filter_coef->pix[12] =  .27641400e-01 ;
        filter_coef->pix[13] = -.25897560e-02 ;
        filter_coef->pix[14] = -.50545260e-02 ;
        filter_coef->pix[15] =  .10501670e-02 ;
#else
#error "The filter length must be one of the known alternatives"
#endif
    return;
}

void free_coef(fgrim *filter_coef){
  fgrim_free(filter_coef);
  return;
}
/******************************************************************************
 *                       filter_periodic_horizontal
 * Goal:
 *   //Apply the half-band low pass filter h0 on each raw of fimg
 *   Apply the filter with coefficients coef on raws of fgin.
 *   The filter is periodic. See test1() for an example.
 * Interface:
 *   fin : [in ] the fgrim to operate on
 *   fout: [out] the result of the filter
 * Return:
 *   status of execution
 ******************************************************************************/
#ifdef UNDEFINED
int filter_periodic_horizontal(fgrim *fg, fgrim *coef, int offset){
  int w, h, raw, cell, f_ind;
  //int offset = 4;
  float cell_val;
  w = fg->w; h = fg->h;
  float * rawbuf = malloc( w * sizeof(float) );
  for (raw=0; raw<h; raw++){ // for each line
    memcpy(rawbuf, &(fg->pix[raw*w]), w*sizeof(float) ); // copy raw in rawbuf
    for (cell=0; cell<w; cell++){ // for each cell, we compute out(cell)
      cell_val = 0;
      for (f_ind=0; f_ind<coef->w; f_ind++){ // for each filter coeff
        cell_val += rawbuf[ (w+cell-offset+f_ind)%w ]* coef->pix[f_ind] ;
        //printf("cell_val(cell=%2d,f_ind=%2d) += (rawbuf[% d]==% 2.2f) * %.0f \n", cell, f_ind, (cell-offset+f_ind)%w, rawbuf[ (cell-offset+f_ind)%w ], coef->pix[f_ind] );
      }
      fg->pix[raw*w+cell] = cell_val;
    }
  }
  free(rawbuf);
  return EXIT_SUCCESS;
}
#else
/*
Optimization :
As in the GPU code, it is faster to do the modulo only if we need it :

    for (cell=0; cell<w; cell++){ // for each cell, we compute out(cell)
      cell_val = 0;
      for (f_ind=0; f_ind<coef->w; f_ind++){ // for each filter coeff
        y = cell-offset+f_ind;         // these two lines give a speedup of 2
        if (y<0 || y>=w) y = (w+y)%w ; // against the previous
        cell_val += fg->pix[raw*w+y]* coef->pix[f_ind] ;
      }
      fgbuf.pix[raw*w+cell] = cell_val;
    }

If this help a lot, we can do better. The if costs too much, and we need it
only at the begining (if(y<0)) and at the end (if(y>w)). Therefore, we separate
the for loop into three for the three different cases.
Again, the speedup is significant (1.5?).
The function below uses these tricks.
*/
int filter_periodic_horizontal(fgrim *fg, fgrim *coef, int offset){
  int w, h, raw, cell, f_ind, y;
  //int offset = 4;
  float cell_val;
  w = fg->w; h = fg->h;
  fgrim fgbuf;
  fgrim_create(&fgbuf, w, h);

  for (raw=0; raw<h; raw++){ // for each line
    // for each cell, we compute out(cell)
    for (cell=0; cell<abs(offset); cell++){
      cell_val = 0;
      for (f_ind=0; f_ind<coef->w; f_ind++){
        y = cell-offset+f_ind;
        if (y<0 || y>=w) y += w ;
        cell_val += fg->pix[raw*w+y]* coef->pix[f_ind] ;
      }
      fgbuf.pix[raw*w+cell] = cell_val;
    }

    for (; cell<w-(offset)-coef->w; cell++){
      cell_val = 0;
      for (f_ind=0; f_ind<coef->w; f_ind++){
        y = cell-offset+f_ind;
        cell_val += fg->pix[raw*w+y]* coef->pix[f_ind] ;
      }
      fgbuf.pix[raw*w+cell] = cell_val;
    }

    for (; cell<w; cell++){
      cell_val = 0;
      for (f_ind=0; f_ind<coef->w; f_ind++){
        y = cell-offset+f_ind;
        if (y>=w) y -= w ;
        cell_val += fg->pix[raw*w+y]* coef->pix[f_ind] ;
      }
      fgbuf.pix[raw*w+cell] = cell_val;
    }


  }

  float * tmp = fg->pix;
  fg->pix = fgbuf.pix;
  fgbuf.pix = tmp;
  fgrim_free(&fgbuf);

  return EXIT_SUCCESS;
}
#endif


/******************************************************************************
 *                       filter_periodic_vertical
 * Goal:
 *   Apply in place the filter with coefficients coef on all columns of fg.
 *   The filter is periodic. See test2() for an example.
 * Interface:
 *   fg: the matrix to operate on
 *   coef: the coef to use fot the filter
 *   nb_of_coef: size of coef.
 * Return:
 *   status of execution
 *****************************************************************************/
#define FILTER_PERIODIC_VERTICAL_VERSION 2

#if ( FILTER_PERIODIC_VERTICAL_VERSION == 1 )
// #warning "FILTER_PERIODIC_VERTICAL_VERSION == 1"
int filter_periodic_vertical(fgrim *fg, fgrim *coef, int offset){
  int w, h, col, cell, f_ind;
  //int offset = 4;
  float cell_val;
  w = fg->w; h = fg->h;
  float * colbuf = malloc( h * sizeof(float) );
  for (col=0; col<w; col++){ // for each column
    for (cell=0; cell<h; cell++) colbuf[cell] = fg->pix[cell*w+col];
    for (cell=0; cell<h; cell++){ // for each cell, we compute out(cell)
      cell_val = 0;
      for (f_ind=0; f_ind<coef->w; f_ind++){ // for each filter coeff
        cell_val += colbuf[ (h+cell-offset+f_ind)%h ]* coef->pix[f_ind] ;
      }
      fg->pix[cell*w+col] = cell_val;
    }
  }
  free(colbuf);
  return EXIT_SUCCESS;
}
#elif ( FILTER_PERIODIC_VERTICAL_VERSION == 2 )
// #warning "FILTER_PERIODIC_VERTICAL_VERSION == 2"
int filter_periodic_vertical(fgrim *fg, fgrim *coef, int offset){
  int w, h, col, cell, f_ind, y;
  //int offset = 4;
  float cell_val;
  w = fg->w; h = fg->h;
  fgrim fgbuf;
  fgrim_create(&fgbuf, w, h);

  for (col=0; col<w; col++){ // for each column

    for (cell=0; cell<abs(offset); cell++){
      cell_val = 0;
      for (f_ind=0; f_ind<coef->w; f_ind++){
        y = cell-offset+f_ind ;
        if (y<0) y += h ; 
        cell_val += fg->pix[y*w+col]* coef->pix[f_ind] ;
      }
      fgbuf.pix[cell*w+col] = cell_val;
    }

    for (; cell<h-(offset)-coef->w; cell++){
      cell_val = 0;
      for (f_ind=0; f_ind<coef->w; f_ind++){
        y = cell-offset+f_ind ;
        cell_val += fg->pix[y*w+col]* coef->pix[f_ind] ;
      }
      fgbuf.pix[cell*w+col] = cell_val;
    }

    for (; cell<h; cell++){
      cell_val = 0;
      for (f_ind=0; f_ind<coef->w; f_ind++){
        y = cell-offset+f_ind ;
        if (y>=h) y -= h ; 
        cell_val += fg->pix[y*w+col]* coef->pix[f_ind] ;
      }
      fgbuf.pix[cell*w+col] = cell_val;
    }

  }

  float * tmp = fg->pix;
  fg->pix = fgbuf.pix;
  fgbuf.pix = tmp;
//  fgrim_free(&fgbuf);

  return EXIT_SUCCESS;
}
#elif ( FILTER_PERIODIC_VERTICAL_VERSION == 3 )
#error "FILTER_PERIODIC_VERTICAL_VERSION 3 NOT IMPLEMENTED YET"
Version 2 is slow because it misses the cache all the time.
Instead, a block approach should ne used.
#else //  ( FILTER_PERIODIC_VERTICAL_VERSION > 3 )
#error "FILTER_PERIODIC_VERTICAL_VERSION should be less than 3"
#endif

/******************************************************************************
 *                       downsample_quincunx_q2
 * Goal:
 *   Downsample image fin using downsampling matrix q2 = [ 1,-1 ; 1,1 ]
 *   Also downsample after a (1,0) time shift
 * Interface:
 *   fin : image to downsample
 *   fout1, fout2: output images
 * Return:
 *   status of execution
 ******************************************************************************/
int downsample_quincunx_q2(fgrim *in, fgrim *outb1, fgrim *outb2){
  int w, h, raw3, col3, raw2, col2, col2b2;
  w = in->w; h = (in->h)/2;
  assert(w == 2*h);
  assert(!(in->h % 2));
  assert( w==outb1->w && w==outb2->w && h==outb1->h && h==outb2->h );  

  for(raw3 = 0; raw3 < h; raw3++){
  for(col3 = 0; col3 < w; col3++){ //both in, outb1, and outb2 have a width of w
    col2 = (2*h + col3 - raw3)%(2*h);
    raw2 = (      col3 + raw3)%(2*h);
    col2b2 = (col2+1) % (2*h);
    //printf("raw3=%d, col3=%d, raw2=%d, col2=%d\n", raw3, col3, raw2, col2);
    outb1->pix[col3 + w*raw3] = in->pix[col2   + w*raw2];
    outb2->pix[col3 + w*raw3] = in->pix[col2b2 + w*raw2];
  }
  }
  return EXIT_SUCCESS;
}

/******************************************************************************
 *                       upsample_quincunx_q2
 * Goal:
 *   Upsample and combine images using downsampling matrix q2 = [ 1,-1 ; 1,1 ]
 * Interface:
 *   inb1, inb2: images to upsample and combine
 *   out: output image
 * Return:
 *   status of execution
 ******************************************************************************/
int upsample_quincunx_q2(fgrim *out, fgrim *inb1, fgrim *inb2){
  int w, h, raw, col, raw_out, col_out1, col_out2;
  w = inb1->w; h=inb1->h;
  assert(w == 2*h);
  assert(w == inb2->w && h == inb2->h);
  assert(w == out->w  && w == out->h );
  for(col=0; col<w; col++){ // columns of input matrices
  for(raw=0; raw<h; raw++){ // raws    of input matrices
    col_out1= (w+ col -raw)%w;
    raw_out = (w+ col +raw)%w;
    //printf("m1=%3d m2=%3d, n1=%3d, n2=%3d\n", col, raw, col_out1, raw_out);
    col_out2= (col_out1+1)%w;
    out->pix[raw_out*w + col_out1] = inb1->pix[raw*w + col]; //  w == width of
    out->pix[raw_out*w + col_out2] = inb2->pix[raw*w + col]; //  the 3 matrices
  }
  }

  return EXIT_SUCCESS;
}



/******************************************************************************
 *                       reorganize_h2v, reorganize_v2h
 * Goal:
 *   h2v: reorganize from an horizontal image 2h=w to a vertical image 2w=h
 *   v2h: reorganize from a vertical image 2w=h to an horizontal image 2h=w
 * Interface:
 *   fin : image to reorganize
 *   fout: reorganized image
 * Return:
 *   status of execution
 ******************************************************************************/
int reorganize_h2v(fgrim *fin, fgrim *fout){
  int wo, ho, i, j;
  wo = fin->h; ho = fin->w;
  assert(2*wo == ho);
  assert( wo==fout->w && ho==fout->h );
  for (i=0; i<wo; i++){ // raw i of fin
    for (j=0; j<wo; j++){ // column j and j+wo of fin
      fout->pix[ i    *wo + j] = fin->pix[i*ho + j   ]; // column j
      fout->pix[(i+wo)*wo + j] = fin->pix[i*ho + j+wo]; // column j+wo
    }
  }
  return EXIT_SUCCESS;
}

int reorganize_v2h(fgrim *fin, fgrim *fout){
  int wo, ho, i, j;
  wo = fin->h; ho = fin->w;
  assert(2*ho == wo);
  assert( wo==fout->w && ho==fout->h );
  for (j=0; j<ho; j++){ // column j of fin
    for (i=0; i<ho; i++){ // raw i and i+ho of fin
      fout->pix[i*wo + j   ] = fin->pix[i     *ho + j]; // raw i
      fout->pix[i*wo + j+ho] = fin->pix[(i+ho)*ho + j]; // raw i+ho
    }
  }
  return EXIT_SUCCESS;
}

/******************************************************************************
 *                       polyphase_combine
 * Goal:
 *   combine matrices such as :
 *   ob1 = ib1 + ib2
 *   ob2 = ib1 - ib2
 * Interface:
 *   ib1, ib2: input matrices. Untouched.
 *   ob1, ob2: output matrices. Must be initially unallocated.
 * Return:
 *   status of execution
 ******************************************************************************/
int polyphase_combine(const fgrim *ib1, const fgrim *ib2, fgrim *ob1,fgrim *ob2)
{
  int w, h, i, j;
  w = ib1->w; h = ib1->h;
  assert(w == ib2->w && h == ib2->h);
  assert(2*w == h);
  for (i=0; i<h; i++){ // raw i
    for (j=0; j<w; j++){ // column j
      ob1->pix[i*w+j] = ib1->pix[i*w+j] + ib2->pix[i*w+j];
      ob2->pix[i*w+j] = ib1->pix[i*w+j] - ib2->pix[i*w+j];
    }
  }
  return EXIT_SUCCESS;
}

/******************************************************************************
 *                       amplify
 * Goal:
 *   multiply every pixel of the matrix by gain
 * Interface:
 *   signal
 *   gain  
 * Return:
 *   status of execution
 ******************************************************************************/
int amplify(fgrim *signal, float gain){
  int w, h, i, j;
  w = signal->w; h = signal->h;
  for (i=0; i<h; i++){ // raw i
    for (j=0; j<w; j++){ // column j
      signal->pix[i*w+j] *= gain;
    }
  }
  return EXIT_SUCCESS;
}

/******************************************************************************
 *                       pair_filter_coef
 * Goal:
 *   create the matching filter pair so that H0(z)==H1(-z)
 * Interface:
 *   lpf: (in ) H0
 *   hpf: (out) H1
 * Return:
 *   status of execution
 ******************************************************************************/
int pair_filter_coef(fgrim *lpf, fgrim *hpf){
  int i, w;
  w = lpf->w;
  fgrim_create(hpf, w, 1);
  for (i=0; i<w-1; i+=2){
    hpf->pix[i  ] =   lpf->pix[i  ];
    hpf->pix[i+1] = - lpf->pix[i+1];
  }
  return EXIT_SUCCESS;
}

/* EOF */
