/*******************************************************************************
 * Author: Jérôme Dubois       Jerome.Dubois@ensimag.imag.fr Jerome@stud.ntnu.no
 * Copyright 2007 2008 Jerome Dubois
 * 
 * 
 ******************************************************************************/

/* To be improved:
 * # fgetc and fputc are used in in xx_write_grim(). Should use fread and
 *   fwrite as in xx_bmp_grim().
 * # There are still some array copy that do not use memcpy.
 */

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <math.h> // link with -lm
#include <assert.h>
#include "grim.h"

/*############################################################################*/
/*#                      BMP structures                                      #*/
/*############################################################################*/
 /* Links to BMP info:
  *   http://en.wikipedia.org/wiki/BMP_file_format
  *   http://www.fortunecity.com/skyscraper/windows/364/bmpffrmt.html
  *   http://www.commentcamarche.net/video/format-bmp.php3 (in French)
  * There is also http://www.martinreddy.net/gfx/ but I've found it too late. It
  *   contains the specifications of many graphic formats and looks good.
  * There already exist loads of code importing/exporting BMPs but they are too
  *   sophisticated for what I want to do (work with grey images only).
  * Hexadecimal: 'cat file | hexdump', xxd, or ghx2 (Gnome, graphic)
  */
#define BMP_FILE_HEADER_SIZE 14
typedef struct /* 14 bytes */
{
    uint8_t b;            /* Must be B */
    uint8_t m;            /* Must be M */
    uint16_t l_size;      /* Size of the file in byte: low and high part */
    uint16_t h_size;      
    uint16_t reserved1;   /* Must be 0 */
    uint16_t reserved2;   /* Must be 0 */
    uint16_t l_offset;    /* Offset from beginning to bitmap data: 1078 bytes */
    uint16_t h_offset;    /*             because 14 + 40 + 4*256 = 1078 */ 
} bmp_file_header ;
/* Notice that the size and offset fields must be separated into two 16-bits
   parts, because the @$%# compiler tries to align the 32-bits types on a word,
   introducing 16 unused bits between m and size. The following relations hold:
   size   = (((uint32_t) h_size)<<16) | ((uint32_t) l_size) ;
   l_size = (uint16_t)  (0x0000FFFF & size) ;
   h_size = (uint16_t) ((0xFFFF0000 & size)>>16) ;
 */

#define BITMAP_INFO_HEADER_SIZE 40 /* DIB = Device Independant Bitmap */
typedef struct /* 40 bytes */ /* Windows V3 */
{
    uint32_t headersize;   /* Must be BITMAP_INFO_HEADER_SIZE */
    int32_t  width;        /* Width  of the image, in pixel */
    int32_t  height;       /* Height of the image, in pixel */
    uint16_t planes;       /* Number of planes. Must be 1 */
    uint16_t depth;        /* Color depth in bits per pixel: 8 */
    uint32_t compression;  /* Type of compression: 0 for non */
    uint32_t imagesize;    /* Size of the image data, in bytes.
                              If no compression, 0 is valid. */
    int32_t hres;          /* Horizontal resolution in pixel per meter */
    int32_t vres;          /* Vertical   resolution in pixel per meter */
    uint32_t nbcolor;      /* nb of colors in the color palette: 0 if 2^depth */
    uint32_t nbimportant;  /* nb of important colors used. 0 if all important */
} bitmap_info_header ;


/*############################################################################*/
/*#                      functions                                           #*/
/*############################################################################*/

/******************************************************************************
 *                       (f)grim_create
 * Goal:
 *   Create the grim img: initialise the size and allocate pixel array
 * Interface:
 *   img   : the grim to create
 *   width : with   of the grim, in pixel
 *   height: height of the grim, in pixel
 * Return:
 *   status of execution
 ******************************************************************************/
int grim_create(grim *img, int width, int height){
  img->w = width;
  img->h = height;
  img->pix = malloc( width * height );
  return EXIT_SUCCESS;
}
int  fgrim_create(fgrim *fimg, int width, int height){
  fimg->w = width;
  fimg->h = height;
  fimg->pix = malloc( width * height * sizeof(float) );
  return EXIT_SUCCESS;
}

/******************************************************************************
 *                       (f)grim_free
 * Goal:
 *   Set size to 0x0 and free pixel array
 * Interface:
 *   img   : the grim to free
 * Return:
 *   status of execution
 ******************************************************************************/
int grim_free(grim *img){
  if (FALSE) printf("grim_free: Doesn't check if img had been allocated\n");
  if (img->w == 0 || img->h == 0) {
    //printf("grim_free: img was already freed it seems...\n");
    return EXIT_SUCCESS;
  }
  img->w = 0;
  img->h = 0;
  free(img->pix);
  return EXIT_SUCCESS;
}
int  fgrim_free(fgrim *fimg){
  if (FALSE) printf("fgrim_free: Doesn't check if img had been allocated\n");
  if (fimg->w == 0 || fimg->h == 0){
    //printf("fgrim_free: img was already freed it seems...\n");
    return EXIT_SUCCESS;
  }
  fimg->w = 0;
  fimg->h = 0;
  free(fimg->pix);
  return EXIT_SUCCESS;
}

/******************************************************************************
 *                       (f)grim_copy
 * Goal:
 *   Copy grim in to grim out
 * Interface:
 *   in : the input  grim to copy
 *   out: the output grim to create and copy to.
 * Return:
 *   status of execution
 ******************************************************************************/
int grim_copy(grim *in, grim *out){
  int i;
  grim_create(out, in->w, in->h) ;
  for (i=0; i<(in->w*in->h); i++) out->pix[i] = in->pix[i] ;
  // TODO: use memcpy
  return EXIT_SUCCESS;
}
int fgrim_copy(fgrim *in, fgrim *out){
  int i;
  fgrim_create(out, in->w, in->h) ;
  for (i=0; i<(in->w*in->h); i++) out->pix[i] = in->pix[i] ;
  // TODO: use memcpy
  return EXIT_SUCCESS;
}

/******************************************************************************
 *                       (f)grim_diff
 * Goal:
 *   Compare grim 1 and 2: Give a hint how much different they are
 * Interface:
 *   im1: grim to compare with im2
 *   im2: grim to compare with im1
 * Return:
 *   The difference in percentage :   0% <-> nearly equals
 *                                  100% <-> one is black and the other white
 *   For the fgrim version, values out of (0,100) can be returned if input
 *   image values are not in the range (0,255)
 * Rq: I could use SQNR comparison method instead. See Wikipedia.
 ******************************************************************************/
int grim_diff(grim *im1, grim *im2){
  int i;
  int diff; // double
  diff=0;
  if ( im1->w!=im2->w || im1->h!=im2->h ) return -1;
  for (i=0; i<(im1->w*im1->h); i++)
    diff += abs( im1->pix[i] - im2->pix[i] ) ;
  return (int)( (float)diff / (float)(im1->w*im1->h) /255.0*100.0 );
}
float fgrim_diff(fgrim *fim1, fgrim *fim2){
  int i;
  float diff; // double
  diff=0;
  if ( fim1->w!=fim2->w || fim1->h!=fim2->h ) return -1;
  for (i=0; i<(fim1->w*fim1->h); i++)
    diff += abs( fim1->pix[i] - fim2->pix[i] ) ;
  // bla: check than it never returnn values bigger than 100 !
 // assert((int)( diff / (fim1->w*fim1->h) /255.0*100.0 ) <= 100);
  return (float)( diff / ( (float)(fim1->w*fim1->h) *255.0 ) *100.0);
}

/******************************************************************************
 *                       grim_cmp
 * Goal:
 *   Compare grim 1 and 2: return true if equals.
 * Interface:
 *   im1: grim to compare with im2
 *   im2: grim to compare with im1
 * Return:
 *   TRUE  if the size and content of the grim are the same
 *   FALSE otherwise
 ******************************************************************************/
int grim_cmp(grim *im1, grim *im2){
  int i;
  if ( im1->w!=im2->w || im1->h!=im2->h ) return FALSE;
  for (i=0; i<(im1->w*im1->h); i++) if (im1->pix[i]!=im2->pix[i]) return FALSE;
  return TRUE;
}
/******************************************************************************
 *                       grim_scale
 * Goal:
 *   Create and initialise a scaled grim of the input grim
 * Interface:
 *   fact: the scaling factor. fact>=1 so out is bigger than in.
 *   in  : the input  grim to scale
 *   out : the output grim to create.
 * Return:
 *   status of execution
 ******************************************************************************/
int grim_scale(int fact, grim *in, grim *out){
  int i,j,k,l;
  grim_create(out, in->w * fact, in->h * fact);
  // TODO: write algo explaination
  for (i=0; i<in->h; i++){
   for (l=0; l<fact;  l++){
    for (j=0; j<in->w; j++){
     for (k=0; k<fact;  k++){
       out->pix[(i*fact+l)*out->w + j*fact+k] = in->pix[j+i*in->w] ;
     }
    }
   }
  }
  return EXIT_SUCCESS;
}
/******************************************************************************
 *                       grim_resize
 * Goal:
 *   Resize GRIM img with the new size.
 *   If new size is bigger, fill with black. If smaller, chop.
 * Interface:
 *   in  : the input  grim to resize
 *   out : the output grim resizeed.
 * Return:
 *   status of execution
 ******************************************************************************/
int grim_resize(grim *in, grim *out, int new_w, int new_h){
  int i,j, w,h, lim_w,lim_h;
  int offset_x, offset_y;
  w = in->w; h = in->h;
#define OFFSET_LOW 10
  offset_x = (new_w - w > OFFSET_LOW) ? OFFSET_LOW : 0;
  offset_y = (new_h - h > OFFSET_LOW) ? OFFSET_LOW : 0;
  lim_w = (w < new_w) ? w : new_w ;
  lim_h = (h < new_h) ? h : new_h ;

  grim_create(out, new_w, new_h);
  grim_set_color(out, 128); // background: medium color

  for (i=0; i< lim_w; i++){ // column i
    for (j=0; j< lim_h; j++){ // raw j
      out->pix[i+offset_x + (j+offset_y)*new_w] = in->pix[i + j*w];
    }
  }

  return EXIT_SUCCESS;
}

/******************************************************************************
 *                       grim_set_color
 * Goal:
 *   Erase a grim with color
 * Interface:
 *   img  : grim to erase
 *   color: color to use
 * Return:
 *   status of execution
 ******************************************************************************/
int grim_set_color(grim *img, bit8 color){
  memset( (void*) img->pix, color, img->w * img->h );
  return EXIT_SUCCESS;
}
int fgrim_set_color(fgrim *fimg, float color){
  memset( (void*) fimg->pix, color, fimg->w * fimg->h * sizeof(float) );
  return EXIT_SUCCESS;
}


/******************************************************************************
 *                       (f)grim_set_sinus
 * Goal:
 *   Draw a sinus in direction direction and with period period
 *   Note that iso-color lines are perpendicular to direction
 * Interface:
 *   (f)img: (f)grim to draw on
 *   color: color to use
 *   direction: angle in radian with the horizontal axis
 *   period: period of the sinus, in pixel
 * Return:
 *   status of execution
 ******************************************************************************/
int grim_set_sinus(grim *img, bit8 color, float direction, int period){
  int i, j;
  double val;
  double Mp_norm, M_norm, teta_M;
/*
  see the figure set_sinus_explain1.{fig,png}
  d   = [angle] direction
  O   = [point] origin: (0,0)
  (d) = [line ] line through O with direction d
  M   = [point] (i,j)
  M'  = [point] orthogonal projection of M on (d)
  teta_M = [angle] angle of point M in polar coordinates = atan(j/i)
  M_norm  = [distance] distance between O and M  = sqrt( i^2 + j^2 )
  Mp_norm = [distance] distance between O and M' = M_norm * cos( teta_M - d )
*/

  for (j=0; j<img->h; j++){ // raw j
    for (i=0; i<img->w; i++){ // column i
      teta_M  =  (0==i) ? PI/2.0 : atan( j / (double)i ) ;
      M_norm  = sqrt( i*i + j*j );
      Mp_norm = M_norm * cos( teta_M - direction );
      val = color /2 * ( sin(Mp_norm *2*PI/period )+1 ) ;
      val = (val>255) ? 255 : val ;
      img->pix[i + j*img->w] = (bit8) val;
    }
  }
#ifdef UNDEFINED
  // tracer la ligne (d) and some well known directions // Not very efficient...
  for (i=0; i<img->w; i++){ // column i
    j =                         /* horitontal */ img->h/2.0 ;                       img->pix[i + j*img->w] = WHITE ;
    j = (int) (i-img->w/2) * tan( 1 * PI/4.0 ) + img->h/2.0 ; if (j>=0 && j<img->h) img->pix[i + j*img->w] = WHITE ;
    j = (int) (i-img->w/2) * tan( 3 * PI/4.0 ) + img->h/2.0 ; if (j>=0 && j<img->h) img->pix[i + j*img->w] = WHITE ;
    j = (int) (i-img->w/2) * tan( 4 * PI/6.0 ) + img->h/2.0 ; if (j>=0 && j<img->h) img->pix[i + j*img->w] = BLACK ;
    j = (int) (i-img->w/2) * tan( 5 * PI/6.0 ) + img->h/2.0 ; if (j>=0 && j<img->h) img->pix[i + j*img->w] = BLACK ;
    j = (int) (i-img->w/2) * tan( 1 * PI/6.0 ) + img->h/2.0 ; if (j>=0 && j<img->h) img->pix[i + j*img->w] = BLACK ;
    j = (int) (i-img->w/2) * tan( 2 * PI/6.0 ) + img->h/2.0 ; if (j>=0 && j<img->h) img->pix[i + j*img->w] = BLACK ;
 // j = (int) (i-img->w/2) * tan( 4 * PI/6.0 ) + img->h/2.0 ; if (j>=0 && j<img->h) img->pix[i + j*img->w] = BLACK ;
 // j = (int) (i-img->w/2) * tan( 5 * PI/6.0 ) + img->h/2.0 ; if (j>=0 && j<img->h) img->pix[i + j*img->w] = BLACK ;
    j = (int) (i-img->w/2) * tan( direction  ) + img->h/2.0 ; if (j>=0 && j<img->h) img->pix[i + j*img->w] = 255 - img->pix[i + j*img->w] ;
  }
  for (j=0; j<img->h; j++){     /* vertical */ i=img->w/2.0 ; /* ...   ...   ... */ img->pix[i + j*img->w] = WHITE ; }
  fprintf(stderr,"Warning: grim_get_sinus also draws well known directions\n");
#endif // (UN)DEFINED
  return EXIT_SUCCESS;
}

int fgrim_set_sinus(fgrim *fimg, bit8 color, float direction, int period){
  int i, j;
  double val;
  double Mp_norm, M_norm, teta_M;
  int w, h;
  w = fimg->w; h = fimg->h;
  
  for (j=0; j<h; j++){ // raw j
    for (i=0; i<w; i++){ // column i
      teta_M  = (0==i) ? PI/2.0 : atan( j / (double)i ) ;
      M_norm  = sqrt( i*i + j*j );
      Mp_norm = M_norm * (cos( teta_M - direction ));
      val = color /2 * ( sin(Mp_norm *2*PI/period )+1 ) ; // fabs(sin(...))
      val = (val>255) ? 255 : val ;
      fimg->pix[i + j*fimg->w] = val;
    }
  }
#ifdef UNDEFINED
  // draw line
  for (i=0; i<w; i++){ // column i
      j = (i-w/2) * tan( direction ) + h/2;
      if (j>=0 && j<h) fimg->pix[i + j*w] = 255; //255 - fimg->pix[i + j*w] ;
  }
#endif // (UN)DEFINED
  return EXIT_SUCCESS;
}

/******************************************************************************
 *                       grim_to_fgrim
 *                       fgrim_to_grim
 * Goal:
 *   Convert a grim to corresponding fgrim or vice versa
 * Interface:
 *   in, fin  : the input  (f)grim to convert
 *   fout, out: the output converted (f)grim. Mustn't be allocated before.
 * Return:
 *   status of execution
 ******************************************************************************/
int grim_to_fgrim(grim *in, fgrim *fout){
  int i;
  fgrim_create(fout, in->w, in->h);
  for (i=0; i<(in->w*in->h); i++) fout->pix[i] = (float) in->pix[i] ;
  return EXIT_SUCCESS;
}
int fgrim_to_grim(fgrim *fin, grim *out){
  int i;
  grim_create(out, fin->w, fin->h);
  for (i=0; i<(fin->w*fin->h); i++) out->pix[i] = (bit8) floor( fin->pix[i] );
  return EXIT_SUCCESS;
}

/******************************************************************************
 *                       (f)grim_write
 * Goal:
 *   Write a (f)grim to stream fpout
 * Interface:
 *   img: the input grim to write
 *   fpout: file pointer to the stream where to write to
 * Return:
 *   status of execution
 ******************************************************************************/
int grim_write(grim *img, FILE *fpout){
  int i,j;
  fprintf(fpout, "%d %d\n", img->w, img->h);
  for (i=0; i<img->h; i++){
    for (j=0; j<img->w; j++){
      fprintf(fpout,"%3d ",img->pix[j + i*img->w]);
    }
    fprintf(fpout,"\n");
  }
  return EXIT_SUCCESS;
}
int fgrim_write(fgrim *fimg, FILE* fpout){
  int i,j;
  fprintf(fpout, "%d %d\n", fimg->w, fimg->h);
    for (i=0; i<fimg->h; i++){ // raw i
      for (j=0; j<fimg->w; j++){ // column j
        fprintf(fpout,"%7.4f ",fimg->pix[j + i*fimg->w]);
      }
      fprintf(fpout,"\n");
  }
  return EXIT_SUCCESS;
}

/******************************************************************************
 *                       (f)grim_print
 * Goal:
 *   Print a (f)grim to stdout
 * Interface:
 *   img: the input grim to print
 * Return:
 *   status of execution
 ******************************************************************************/
int grim_print(grim *img){
  return grim_write(img, stdout);
}
int fgrim_print(fgrim *fimg){
  return fgrim_write(fimg, stdout);
}

/******************************************************************************
 *                       (f)grim_write_bmp
 * Goal:
 *   Write a grim to a BMP file
 * Interface:
 *   img: the input grim to write
 *   bmp: name of BMP file to write to. It should have the .bmp extention
 * Return:
 *   status of execution
 ******************************************************************************/
int grim_write_bmp(grim *img, char* bmp){
  FILE *fpout;
  int i;
  uint32_t size, offset;
  bmp_file_header    fh;
  bitmap_info_header ih;
  
  fpout = fopen(bmp,"wb");
  if (!fpout) {
    fprintf(stderr, "[ERROR] Cannot open bmp file\n");
    exit(-1);
  }
  
  /* File header */
  fh.b         =  'B';
  fh.m         =  'M';
  size         = /* Size of the file */
                 sizeof(bmp_file_header)
               + sizeof(bitmap_info_header)
               + 4 * 256          /* palette */
               + img->w * img->h; /* pixel array */
  fh.l_size = (uint16_t)  (0x0000FFFF & size) ;
  fh.h_size = (uint16_t) ((0xFFFF0000 & size)>>16) ;
  fh.reserved1 =    0;
  fh.reserved2 =    0;
  offset       = (int32_t ) 1078;
  fh.l_offset  = (uint16_t)  (0x0000FFFF & offset) ;;
  fh.h_offset  = (uint16_t) ((0xFFFF0000 & offset)>>16) ;
  fwrite( &fh, sizeof(bmp_file_header), 1, fpout);
  if ( sizeof(bmp_file_header) != BMP_FILE_HEADER_SIZE ){
    fprintf(stderr,"Error: The file header size is wrong\n");
  }
  
  /* Bitmap Information (DIB header) */
  ih.headersize = 40;
  ih.width      = img->w;
  ih.height     = img->h;
  ih.planes     = 1;
  ih.depth      = 8;    /* bits per pixel */
  ih.compression= 0;    /* no compression */
  ih.imagesize  = img->w * img->h; /* bytes */ /* 0 accepted: no compression */
  ih.hres       = 2835; /* pixel per meter */  /* 0 accepted */
  ih.vres       = 2835; /* pixel per meter */  /* 0 accepted */
  ih.nbcolor    = 256;                         /* 0 accepted: default 2^depth */
  ih.nbimportant= 256;                         /* 0 accepted: all important */
  fwrite( &ih, sizeof(bitmap_info_header), 1, fpout);
  
  /* Color Palette */
  for (i=0; i<256; i++){
    fputc( (char) i, fpout) ;
    fputc( (char) i, fpout) ;
    fputc( (char) i, fpout) ;
    fputc( (char) 0, fpout) ;
  }
  
  /* Pixel array */
  fwrite( img->pix, img->w*img->h, 1, fpout);
  // same as: for (i=0; i<(img->w*img->h); i++) fputc( (char)img->pix[i], fpout) ;
  
  fclose(fpout);
  return EXIT_SUCCESS;
}

int fgrim_write_bmp(fgrim *fimg, char* bmp){
  grim img;
  fgrim_to_grim(fimg, &img);
  grim_write_bmp(&img, bmp);
  grim_free(&img);
  return EXIT_SUCCESS;
}


/******************************************************************************
 *                       grim_read_bmp
 * Goal:
 *   Write a grim to a BMP file
 * Interface:
 *   img: the input grim to write
 *   bmp: name of BMP file to write to. It should have the .bmp extention
 * Return:
 *   status of execution
 ******************************************************************************/
int grim_read_bmp(char* bmp, grim *img){
  FILE *fpin;
  uint32_t size, offset;
  bmp_file_header    fh;
  bitmap_info_header ih;

  /* open input file */
  fpin = fopen(bmp,"rb");
  if (!fpin) {
    fprintf(stderr, "[ERROR] Cannot open bmp file\n");
    exit(-1);
  }
  
  /* get headers */
  fread( &fh, sizeof(bmp_file_header   ), 1, fpin);
  fread( &ih, sizeof(bitmap_info_header), 1, fpin);

  /* Some checks could be done here, at several levels: check that the file
   * begins with "BM", that it is well formed (real size == size announced, the
   * zeros are zeros,...), that it uses a 256 colors palette,
   * that the palette is 0000 1110 2220 3330 ... 256,256,256,0
   * GIMP exported images are OK. Not critical anyway.
   */
   
  /* file header checks */
  assert(fh.b=='B');
  assert(fh.m=='M');
  size   = (((uint32_t) fh.h_size  )<<16) | ((uint32_t) fh.l_size  ) ;  
  assert(fh.reserved1 == 0);
  assert(fh.reserved2 == 0);
  offset = (((uint32_t) fh.h_offset)<<16) | ((uint32_t) fh.l_offset) ;
  assert(offset == 1078);
  
  /* Bitmap info header checks */
  assert(ih.headersize == BITMAP_INFO_HEADER_SIZE);
  assert(ih.planes == 1);
  assert(ih.depth == 8);
  assert(ih.compression == 0);
  assert(ih.imagesize == 0 || ih.imagesize == ih.width * ih.height);
//  assert(ih.hres == 0 || ih.hres == 2835);
//  assert(ih.hres == 0 || ih.hres == 2835);
  assert(ih.nbcolor == 0 || ih.nbcolor == 256);
  assert(ih.nbimportant == 0 || ih.nbimportant == 256);
  
  /* palette ckeck */
  // NOT IMPORTANT
  
  /* copy image data */
  grim_create(img, ih.width, ih.height);
  fseek(fpin, offset, SEEK_SET);
  fread(img->pix, img->w*img->h, 1, fpin);
  
  fclose(fpin);
  return EXIT_SUCCESS;
}

/******************************************************************************
 *                       grim_write_grim
 * Goal:
 *   Write a grim to a GRIM file
 * Interface:
 *   imgin: the input grim to write
 *   fileout: name of GRIM file to write to. It should have the .grim extention
 * Return:
 *   status of execution
 ******************************************************************************/
int grim_write_grim(grim *imgin, char* fileout){
  FILE *fpout;
  int i;
  unsigned int j, k;
  char c;
  
  fpout = fopen(fileout,"wb");
  if (!fpout) {
    fprintf(stderr, "[ERROR] Cannot open grim file\n");
    exit(-1);
  }
  
  j = (unsigned int) imgin->w;
  k = j & 0x000000FF; c= (char) (k>>0)   ; fputc(c,fpout) ;
  k = j & 0x0000FF00; c= (char) (k>>8)   ; fputc(c,fpout) ;
  k = j & 0x00FF0000; c= (char) (k>>16)  ; fputc(c,fpout) ;
  k = j & 0xFF000000; c= (char) (k>>24)  ; fputc(c,fpout) ;
  j = (unsigned int) imgin->h;
  k = j & 0x000000FF; c= (char) (k>>0)   ; fputc(c,fpout) ;
  k = j & 0x0000FF00; c= (char) (k>>8)   ; fputc(c,fpout) ;
  k = j & 0x00FF0000; c= (char) (k>>16)  ; fputc(c,fpout) ;
  k = j & 0xFF000000; c= (char) (k>>24)  ; fputc(c,fpout) ;
  
  for (i=0; i<(imgin->w*imgin->h); i++) fputc( (char)imgin->pix[i], fpout) ;
  // VERY BAD: should use fwrite
  // fwrite(imgin->pix, imgin->h * imgin->w, 1, fpout); // NOT TESTED
  
  fclose(fpout);
  return EXIT_SUCCESS;
}

/******************************************************************************
 *                       grim_read_grim
 * Goal:
 *   Read a GRIM file and copy into imgout
 * Interface:
 *   filein: name of GRIM file to read from. It should have the .grim extention
 *   imgout: the grim to write in
 * Return:
 *   status of execution
 ******************************************************************************/
int grim_read_grim(char* filein, grim *imgout){
  FILE *fpin;
  int i, w, h;
  unsigned int u;
  char c;
  
  fpin = fopen(filein,"rb");
  if (!fpin) {
    fprintf(stderr, "[ERROR] Cannot open grim file\n");
    exit(-1);
  }
  
  c = fgetc(fpin); u  = 0x000000FF &    (unsigned int) c       ;
  c = fgetc(fpin); u |= 0x0000FF00 & ( ((unsigned int) c)<<8  );
  c = fgetc(fpin); u |= 0x00FF0000 & ( ((unsigned int) c)<<16 );
  c = fgetc(fpin); u |= 0xFF000000 & ( ((unsigned int) c)<<24 );
  w = u;
  c = fgetc(fpin); u  = 0x000000FF &    (unsigned int) c       ;
  c = fgetc(fpin); u |= 0x0000FF00 & ( ((unsigned int) c)<<8  );
  c = fgetc(fpin); u |= 0x00FF0000 & ( ((unsigned int) c)<<16 );
  c = fgetc(fpin); u |= 0xFF000000 & ( ((unsigned int) c)<<24 );
  h = u;

  grim_create(imgout, w, h);
  for (i=0; i<(imgout->w*imgout->h); i++) imgout->pix[i] = (bit8)fgetc(fpin);
  // VERY BAD: should use fread

  fclose(fpin);
  return EXIT_SUCCESS;
}

/******************************************************************************
 *                       grim_text_read
 * Goal:
 *   Read a text file representing a grim, and copy into imgout
 * Interface:
 *   filein: name of text file to read from.
 *   imgout: the grim to write in
 * Return:
 *   status of execution
 ******************************************************************************/
int grim_text_read(char* filein, grim *imgout){
  FILE *fpin;
  int w, h, raw, col;
  int tmp;
  
  /* open input file */
  fpin = fopen(filein,"r");
  if (!fpin) {
    fprintf(stderr, "[ERROR] Cannot open grtxt file\n");
    exit(-1);
  }

  /* read width and height */
  fscanf(fpin, "%d %d\n", &w, &h);
  fprintf(stdout, "read from text file: w=%d h=%d\n", w, h);
  grim_create(imgout, w, h);
  
  /* read and fill image */
  for (raw=0; raw<h; raw++){
	  for (col=0; col<w-1; col++){
	      fscanf(fpin, "%d", &tmp);
	      imgout->pix[raw*w + col]=tmp;
	  }
      fscanf(fpin, "%d\n", &tmp );
      imgout->pix[raw*w + col]=tmp;
  }

  return EXIT_SUCCESS;
}

/******************************************************************************
 *                       (f)grim_line_to_2d
 * Goal:
 *   draw a 2D plot of line <line> of image <fin>
 * Interface:
 *   fin: image whose <line>th line must be used as ordinates
 *   line: number of the line of <fin> to use to plot
 *   out_2d: output image of size (fin->w,256) representing plot (i,fin->pix[i])
 * Return:
 *   status of execution
 ******************************************************************************/
int  grim_line_to_2d(grim  *fin, int line, grim *out_2d){
  int w, h, i;
  w = fin->w;
  h = 256;
  grim_create(out_2d, w, h);
  grim_set_color(out_2d, WHITE);
  for (i=0; i<w; i++){ out_2d->pix[ i + w * fin->pix[i] ] = BLACK; }
  return EXIT_SUCCESS;
}
int fgrim_line_to_2d(fgrim *fin, int line, grim *out_2d){
  int i, w, h, h2, val, extent;
  w = fin->w;
  h2 = 256;
  extent = 50; // we accept values at max extent pixels out of the [0,255] range
  h = h2 + 2*extent; // nb of values kept
  grim_create(out_2d, w, h);
  grim_set_color(out_2d, WHITE);
  for (i=0; i<w; i++){ out_2d->pix[i + w*extent     ] = 200; } // draw y = 0
  for (i=0; i<w; i++){ out_2d->pix[i + w*(h2+extent)] = 200; } // draw y = h2
  
  // transform it from float to bit8.
  for (i=0; i<w; i++){
    val = (int) floor( fin->pix[i+line*w] );
    val = (val>h2-1+extent) ? h2-1+extent : val;
    val = (val< 0  -extent) ?  0  -extent : val;
    out_2d->pix[ i + w * (val+extent) ] = BLACK;
  }
  
/*  ret = grim_line_to_2d(&linebuf, 0, out_2d);
  grim_free(&linebuf);*/
  return EXIT_SUCCESS;
}

// EOF
/*******************************************************************************
i = snprintf(bmp_file_header, BMP_FILE_HEADER_SIZE, "BM%d%d%d", size, 0, 1078 );
if (i < 0 || i >= BMP_FILE_HEADER_SIZE)
                            fprintf(stderr,"Error: grim_write_bmp: snprintf\n");
*/
