/**
 * File  : grim2x.c
 * Author: Jérôme Dubois <Jerome.Dubois@ensimag.imag.fr>
 * Date  : 2007-12-20
 * Copyright 2007 Jerome Dubois
 *
 * Description:
    Convert grey image in different formats :
     * GRIM  image.grim
     * BMP   image.bmp
     * text  image.grtxt
 * Format:
    grim2x [-h] [-o <image_out>] <conv> <image_in>
     -h :        print help and exit
     <conv>      the conversion to make: one of "grim2bmp", "bmp2grim",
                 "grim2grtxt", "grtext2grim", "grim_print", "bmp_print"
     <image_in>  is the name of input image to convert
     <image_out> is the name of output converted image.
                 Default is out.grim, out.grtxt, out.bmp, or stdout
 */
 /********************************************************************************/
#include <stdio.h>     /* for printf */
#include <stdlib.h>    /* for exit  */
#include <assert.h>    /* for assert */
#include <string.h> 
#include <getopt.h> 

#include "grim.h"

#define ARG_STR_SIZE 256   // max size of string arguments (including '\0')

#define NOTASK     0
#define GRIM2BMP   1
#define BMP2GRIM   2
#define GRIM2GRTXT 3
#define GRTXT2GRIM 4
#define GRIM_PRINT 5
#define BMP_PRINT  6

/*******************************************************************************
 * Print the help message  
 * command: used (argv[0])char * command
 ******************************************************************************/
void print_help(char * command) {
    printf("NAME\n"                                                           );
    printf("      %s\n",command                                               );
    printf("\nSYNOPSIS\n"                                                     );
    printf("      %s [-h] [-o <image_out>] <conv> <image_in>\n",command       );
    printf("\nDESCRIPTION\n"                                                  ); 
    printf("      -h :print help and exit\n"                                  );
    printf("      <conv>: one of grim2bmp, bmp2grim, grim2grtxt,\n"           );
    printf("                     grtext2grim, grim_print, bmp_print"          );
    printf("\nAUTHOR\n"                                                       );
    printf("      Jerome Dubois <Jerome.Dubois@ensimag.imag.fr>\n"            );
    printf("\nCOPYRIGHT\n"                                                    );
    printf("      Copyright (c) Jérome Dubois\n"                              );
}

/*******************************************************************************
 * print error message on stderr 
 * error_message: message to print 
 * command:       command used (argv[0])
 ******************************************************************************/
void report_error(char * command, char * error_message) {
    fprintf(stderr, "[ERROR] %s\n",error_message);
    fprintf(stderr, "Use '%s -h' for help\n", command);
    exit(EXIT_FAILURE);
}

/*******************************************************************************
 * main
 ******************************************************************************/
int main(int argc, char* argv[]){

    // Command line management 
    extern char *optarg;  // specific to getopt
    extern int optind;    //, opterr, optopt;  // id.
    char c;
    char conv[ARG_STR_SIZE];
    int  task=NOTASK;
    char image_in[ARG_STR_SIZE];
    char image_out[ARG_STR_SIZE];
    int  option_o = FALSE ;
    grim img;
    
    // Management of parameters
    while ((c = getopt(argc, argv, "ho:")) != -1) {
        switch(c) {
        case 'h': print_help(argv[0]); return EXIT_SUCCESS;
        case 'o':
            option_o = TRUE;
            assert(strlen(optarg) < ARG_STR_SIZE); // check size
            strncpy(image_out,optarg,ARG_STR_SIZE);
            if (image_out[0] == '-') report_error(argv[0], "Bad argument!");
            break;
        default: report_error(argv[0], "Bad argument!");
        }
    }
    
    // parameter: conv
    if ((argc - optind) != 2) report_error(argv[0], "Bad argument!");
    assert(strlen(argv[optind]) < ARG_STR_SIZE);
    strncpy(conv,argv[optind],ARG_STR_SIZE);
    if      (!strncmp(conv, "grim2bmp"  , ARG_STR_SIZE))
                   { task = GRIM2BMP   ; if (!option_o) strncpy(image_out, "out.bmp"  , ARG_STR_SIZE) ; }
    else if (!strncmp(conv, "bmp2grim"  , ARG_STR_SIZE))
                   { task = BMP2GRIM   ; if (!option_o) strncpy(image_out, "out.grim" , ARG_STR_SIZE) ; }
    else if (!strncmp(conv, "grim2grtxt", ARG_STR_SIZE))
                   { task = GRIM2GRTXT ; if (!option_o) strncpy(image_out, "out.grtxt", ARG_STR_SIZE) ; }
    else if (!strncmp(conv, "grtxt2grim", ARG_STR_SIZE))
                   { task = GRTXT2GRIM ; if (!option_o) strncpy(image_out, "out.grim" , ARG_STR_SIZE) ; }
    else if (!strncmp(conv, "grim_print", ARG_STR_SIZE))
                   { task = GRIM_PRINT ; }
    else if (!strncmp(conv, "bmp_print" , ARG_STR_SIZE))
                   { task = BMP_PRINT  ; }
    else report_error(argv[0], "Bad argument!");

     // parameter: image_in
    optind++;
    assert(strlen(argv[optind]) < ARG_STR_SIZE);
    strncpy(image_in,argv[optind],ARG_STR_SIZE);
    
    // verification
#ifdef DEBUB
    fprintf(stdout,"Arg: task=%d, conv=|%s|\n",task, conv);
    fprintf(stdout,"Arg: option_o: %s\n", option_o ? "yes":"no");
    fprintf(stdout,"Arg: image_in =|%s|\n",image_in);
    fprintf(stdout,"Arg: image_out=|%s|\n",image_out);
#endif // #define DEBUG

    // execution
    switch(task){
      case NOTASK    : report_error(argv[0], "no task"); break;
      case GRIM2BMP  : 
        grim_read_grim(image_in,&img);
        grim_write_bmp(&img,image_out);
        grim_free(&img);
        break;
      case BMP2GRIM  :
        grim_read_bmp(image_in,&img);
        grim_write_grim(&img,image_out);
        grim_free(&img);
        break;
      case GRIM2GRTXT:
        grim_read_grim(image_in,&img);
        grim_write(&img,stdout);           // THIS IS NOT THE EXPECTED BEHAVIOUR
        grim_free(&img);
        break;
      case GRTXT2GRIM: 
        grim_text_read(image_in,&img);
        grim_write_grim(&img,image_out);
        break;
      case GRIM_PRINT:
        grim_read_grim(image_in,&img);
        grim_print(&img);
        grim_free(&img);
        break;
      case BMP_PRINT :
        grim_read_bmp(image_in,&img);
        grim_print(&img);
        grim_free(&img);
        break;
      default: report_error(argv[0], "Bad task");
    }
    
    return EXIT_SUCCESS;
}

// -EOF-
