/**
 * File   txt2grim.c
 * Author Jérôme Dubois <Jerome.Dubois@ensimag.imag.fr>
 * Date   2007-12-14
 * Copyright 2007 Jerome Dubois
 *
 * Description
    create a GRIM image from a TXT image. See grim.h for more information.  
   ./txt2grim [-o grim_image] [-h] txt_image
     txt_image : input image to transform into GRIM.
     grim_image  : output GRIM image to write to. If it exists, it is overwritten.
                  default name is "out.grim"
                  (should be input name but with .grim extention.)
     -h : print help
 */
 /********************************************************************************/
#include <stdio.h>     /* for printf */
#include <stdlib.h>    /* for exit  */
#include <assert.h>    /* for assert */
#include <string.h> 
#include <getopt.h> 

#include "grim.h"

#define STRING_SIZE 256   // maximal size of strings, including '\0'

void printHelp(char * command);                        // print help message
void printError(char * error_message, char * command); // print error message

int main(int argc, char * argv[]){

    // Command line management 
    extern char *optarg;  // specific to getopt
    extern int optind;    //, opterr, optopt;  // id.
    char c;
    char arg_o[STRING_SIZE]="out.grim";
    char arg_i[STRING_SIZE];
    int  option_o = 0 ;
 
    // Management of parameters in command line
    while ((c = getopt(argc, argv, "o:h")) != -1) {
        switch(c) {
        case 'h': printHelp(argv[0]); return EXIT_SUCCESS;
        case 'o':
            option_o = 1;
            assert(strlen(optarg) < STRING_SIZE); // check size
            strncpy(arg_o,optarg,STRING_SIZE);
            if (arg_o[0] == '-') printError("Bad Format for arguments!",argv[0]);
            break;
        default: printError("Bad Format!",argv[0]);
        }
    }
    
    // parameter: arg_i
    if ((argc - optind) != 1) printError("Bad Format",argv[0]);
    assert(strlen(argv[optind]) < STRING_SIZE);
    strncpy(arg_i,argv[optind],STRING_SIZE);
    
    // parameter: arg_o
    if (!option_o) /* set default */;
    
    // verification
    //fprintf(stdout,"Valeur de arg_o: %s\n",arg_o);
    //fprintf(stdout,"Valeur de arg_i: %s\n",arg_i);
    
    // execution
    grim img;
    grim_text_read(arg_i,&img);
    fprintf(stdout,"txt2grim: After read, image sizes are: w=%d, h=%d\n", img.w, img.h);
    grim_write_grim(&img,arg_o);
    
/*  // begin debug
    grim img2, img3;
    grim_create(&img3, 130, 131);
    grim_set_color(&img3, 133);
    grim_write_grim(&img3,"img3.grim");
    grim_read_grim("img3.grim",&img2);
    fprintf(stdout,"txt2grim: img3: w=%d, h=%d\n", img3.w, img3.h);
    fprintf(stdout,"txt2grim: img2: w=%d, h=%d\n", img2.w, img2.h);
    grim_free(&img2);
    grim_free(&img3);
*/  // end debug

    grim_free(&img);    
    return EXIT_SUCCESS;
}

/** 
 * Print the help message  
 * param: command used (argv[0])
 */
void printHelp(char * command) {
    printf("NAME\n"                                                           );
    printf("      %s\n",command                                               );
    printf("\nSYNOPSIS\n"                                                     ); 
    printf("      %s [-h]\n",command                                          );
    printf("      %s [-o grim_image] txt_image\n",command                     );
    printf("\nDESCRIPTION\n"                                                  ); 
    printf("      -h : print help and exit\n"                                  );
    printf("      -o grim_image: a grim file name\n"                            );
    printf("         Default value: \"out.grim\"\n"                           );
    printf("\nAUTHOR\n"                                                       );
    printf("      Jerome Dubois <Jerome.Dubois@ensimag.imag.fr>\n"            );
    printf("\nCOPYRIGHT\n"                                                    );
    printf("      Copyright (c) Jérome Dubois\n"                              );
}

/**
 * print error message on stderr 
 * param error_message: Error message to print 
 * param command      : command used (argv[0])
 */
void printError(char * error_message, char * command) {
    fprintf(stderr, "[ERROR] %s\n",error_message);
    fprintf(stderr, "Use '%s -h' for help\n", command);
    exit(EXIT_FAILURE);
}
